from collections import defaultdict
from typing import Callable, Dict, Iterable, List, Mapping, Optional

import torch
from countergen.types import AugmentedSample, Category
from countergentorch.tools.utils import get_gpt_tokenizer
from torch import nn
from transformers import BatchEncoding, GPT2LMHeadModel


def get_mlp_modules(model: GPT2LMHeadModel, layer_numbers: Optional[List[int]]) -> Dict[str, nn.Module]:
    model_transformer: nn.ModuleList = model.transformer.h  # type: ignore
    layer_numbers_ = layer_numbers or list(range(len(model_transformer)))
    names = [f"transformer.h.{n}.mlp" for n in layer_numbers_]
    return {name: model.get_submodule(name) for name in names}  # type: ignore


def get_res_modules(model: GPT2LMHeadModel, layer_numbers: Optional[List[int]]) -> Dict[str, nn.Module]:
    model_transformer: nn.ModuleList = model.transformer.h  # type: ignore
    layer_numbers_ = layer_numbers or list(range(len(model_transformer)))
    names = [f"transformer.h.{n}" for n in layer_numbers_]
    return {name: model.get_submodule(name) for name in names}  # type: ignore


def get_corresponding_activations(
    samples: Iterable[AugmentedSample], model: nn.Module, modules: Iterable[nn.Module]
) -> Mapping[Category, List[Dict[nn.Module, torch.Tensor]]]:
    """For each category, returns a list of activations obtained by running the variations corresponding to this category."""

    tokenizer = get_gpt_tokenizer()

    operation = lambda t: t.reshape((-1, t.shape[-1]))

    activations_by_cat = defaultdict(lambda: [])
    for sample in samples:
        for variations in sample.get_variations():
            acts = get_activations(tokenizer(variations.text, return_tensors="pt"), model, modules, operation)
            for cat in variations.categories:
                activations_by_cat[cat].append(acts)
    return activations_by_cat


Operation = Callable[[torch.Tensor], torch.Tensor]


def get_activations(
    tokens: BatchEncoding, model: nn.Module, modules: Iterable[nn.Module], operation: Operation = lambda x: x
) -> Dict[nn.Module, torch.Tensor]:
    handles = []
    activations: Dict[nn.Module, torch.Tensor] = {}

    def hook_fn(module, inp, out):
        out_ = out[0] if isinstance(out, tuple) else out

        activations[module] = operation(out_.detach())

    for module in modules:
        handles.append(module.register_forward_hook(hook_fn))
    try:
        model(**tokens.to(model.device))
    except Exception as e:
        raise e
    finally:
        for handle in handles:
            handle.remove()
    return activations


# (module, input, output) -> output
ModificationFn = Callable[[nn.Module, torch.Tensor, torch.Tensor], torch.Tensor]


def run_and_modify(
    tokens: BatchEncoding, model: nn.Module, modification_fns: Dict[nn.Module, ModificationFn] = {}
) -> BatchEncoding:
    handles = []
    for module, f in modification_fns.items():
        handles.append(module.register_forward_hook(f))  # type: ignore
    try:
        out = model(**tokens.to(model.device))
        return out
    except Exception as e:
        raise e
    finally:
        for handle in handles:
            handle.remove()



from typing import Any, Dict, Optional, Tuple
import torch
from torch import nn

import countergen.config
from countergen.tools.utils import maybe_tqdm


def get_bottlenecked_linear(input_dim: int, output_dim: int, bottleneck_dim: int = 1):
    return nn.Sequential(nn.Linear(input_dim, bottleneck_dim), nn.Linear(bottleneck_dim, output_dim))


def get_bottlenecked_mlp(input_dim: int, output_dim: int, bottleneck_dim: int = 8, hidden_dim: int = 64):
    return nn.Sequential(
        nn.Linear(input_dim, bottleneck_dim),
        nn.Linear(bottleneck_dim, hidden_dim),
        nn.ReLU(),
        torch.nn.Linear(hidden_dim, output_dim),
    )


def fit_model(
    model: nn.Module,
    ds: torch.utils.data.Dataset,
    max_iters: int,
    loss_fn: nn.Module = nn.CrossEntropyLoss(),
    adam_kwargs: Dict[str, Any] = {"lr": 1e-4},
    dataloader_kwargs: Dict[str, Any] = {"batch_size": 256, "shuffle": True},
    print_pbar: Optional[bool] = None,
) -> Dict[str, float]:
    optimizer = torch.optim.Adam(model.parameters(), **adam_kwargs)
    dataloader = torch.utils.data.DataLoader(ds, **dataloader_kwargs)

    print_pbar = print_pbar or countergen.config.verbose >= 3
    tepoch = maybe_tqdm(range(max_iters), print_pbar)

    for _ in tepoch:
        epoch_loss = 0.0
        n_correct = 0
        n_tot = 0

        for x, y in dataloader:
            optimizer.zero_grad()
            out = model(x)

            loss_val = loss_fn(out, y)
            epoch_loss += loss_val.item()
            loss_val.backward()
            optimizer.step()

            with torch.no_grad():
                preds = torch.argmax(out, dim=-1)
                n_correct += (preds == y).sum().item()
                n_tot += len(preds)

        if print_pbar >= 3:
            tepoch.set_postfix(loss=epoch_loss, accuracy=n_correct / n_tot)  # type:ignore

    return {"loss": epoch_loss, "accuracy": n_correct / n_tot}



from math import exp
from typing import TYPE_CHECKING, Callable, Sequence, List, Optional, Tuple

import torch
from countergen.tools.math_utils import perplexity
from countergentorch.tools.utils import concat_dicts, get_device, get_gpt_tokenizer, remove_last_tok
from countergen.types import Input, ModelEvaluator, Outputs, Performance
from transformers import BatchEncoding, GPT2LMHeadModel, GPT2Tokenizer

metrics = ["perplexity", "probability"]

LogProbs = float

# Return a log prob for each token of output
GenerativeModel = Callable[[Input, Outputs], Sequence[Sequence[LogProbs]]]


def pt_to_generative_model(model: torch.nn.Module, device: Optional[str] = None) -> GenerativeModel:
    """Make a GenerativeModel out of a pytorch model.

    The model should take {"input_ids": [tensor], "attention_mask": [tensor]} as input,
    and return something that has a "logits" attribute."""

    tokenizer = get_gpt_tokenizer()
    device = device or get_device()
    model = model.to(device)

    def gen_model(inp: Input, out: Outputs) -> List[List[float]]:
        tokens_inp = tokenizer(inp, return_tensors="pt").to(device)
        token_outs = [tokenizer(o, return_tensors="pt").to(device) for o in out]

        return get_correct_logprobs(tokens_inp, token_outs, model)

    return gen_model


def get_generative_model_evaluator(model: GenerativeModel, metric: str = "probability") -> ModelEvaluator:
    """Return the ModelEvaluator corresponding to the model & the metric.

    Available metrics: probability & perplexity"""

    def run(inp: Input, out: Outputs) -> Performance:
        if len(out) == 0:
            raise ValueError("Expected output should be provided for gpt models")
        if len(inp) == 0:
            raise ValueError("Empty inputs are forbidden for gpt models.")

        correct_log_probs_list = model(inp, out)

        total_prob: float = 0
        total_log_prob: float = 0
        number_of_toks: int = 0
        for correct_log_probs in correct_log_probs_list:
            total_prob += exp(sum(correct_log_probs))
            number_of_toks += len(correct_log_probs)
            total_log_prob += sum(correct_log_probs)

        if metric == "perplexity":
            return exp(-total_log_prob / number_of_toks)
        if metric == "probability":
            return total_prob
        raise ValueError(f"{metric} is not a valid metric. Choose one in {metrics}.")

    return run


def get_correct_logprobs(
    tokens_inp: BatchEncoding, token_outs: List[BatchEncoding], model: torch.nn.Module
) -> List[List[float]]:

    if all([o["input_ids"].shape[-1] == 1 for o in token_outs]):
        return [[x.item() for x in t] for t in get_correct_1tok_logprobs(tokens_inp, token_outs, model)]

    inp_length = tokens_inp["input_ids"].shape[-1]

    result: List[List[float]] = []

    for tokens_out in token_outs:
        out_length = tokens_out["input_ids"].shape[-1]
        assert out_length > 0, "Zero length expected output is forbidden"

        tokens_to_feed = remove_last_tok(concat_dicts([tokens_inp, tokens_out]))
        with torch.no_grad():
            logits = model(**tokens_to_feed).logits[0]
        log_probs = torch.log_softmax(logits, dim=-1)[inp_length - 1 :, :]

        assert len(log_probs) == len(tokens_out["input_ids"][0])
        correct_log_probs = torch.gather(log_probs.to("cpu"), 1, tokens_out["input_ids"][0, :, None].to("cpu"))[:, 0]

        result.append([x.item() for x in correct_log_probs])

    return result


def get_correct_1tok_logprobs(
    tokens_inp: BatchEncoding, token_outs: List[BatchEncoding], model: torch.nn.Module
) -> List[torch.Tensor]:

    with torch.no_grad():
        logits = model(**tokens_inp).logits[0].to("cpu")
    log_probs = torch.log_softmax(logits, dim=-1)[-1:, :]

    good_tokens = [o["input_ids"][0, 0].item() for o in token_outs]

    correct_log_probs = [log_probs[:, i] for i in good_tokens]

    return correct_log_probs



from typing import Optional, List

import torch
import countergen.config
from countergen.types import Input, ModelEvaluator, Outputs, Performance
from torch import nn
from transformers import Pipeline


def get_classification_pipline_evaluator(pipeline: Pipeline) -> ModelEvaluator:
    """Returns a function which evaluate the pipeline on a (input,output) pair.

    The output of the pipeline must contain a "label" field, which is the prediction.
    The function returns 1 if the prediction matches the ouput and 0 otherwise."""

    def run(inp: Input, out: Outputs) -> Performance:
        assert len(out) == 1, "There should be only one correct label"
        true_label = out[0]

        pred = pipeline(inp)[0]
        if "label" not in pred:
            raise ValueError(f"pipeline shoud ouput a dict containing a label field but pred={pred}")
        perf = 1.0 if true_label == pred["label"] else 0.0
        if countergen.config.verbose >= 4:
            print(f"inp={inp} true_label={true_label} pred={pred} perf={perf}")
        return perf

    return run


def get_classification_model_evaluator(
    model: nn.Module, tokenizer, labels: List[str], metric: str = "correct_prob"
) -> ModelEvaluator:
    """Returns a function which evaluate the model on a (input,output) pair.

    The tokenizer will be called using __call__, and must support the return_tensors="pt" argument.
    The output of the model must contain a "logits" field, which is the prediction logits.

    if metric="correct_prob", returns the probability of the correct token.
    if metric="accuracy", returns 1 if the top-1 prediction matches the ouput and 0 otherwise.
    otherwise returns ValueError."""

    def run(inp: Input, out: Outputs) -> Performance:
        assert len(out) == 1, "There should be only one correct label"
        true_label = out[0]

        tokens = tokenizer(inp, return_tensors="pt")
        with torch.no_grad():
            pred_logits = model(**tokens).logits[0]

        if metric == "accuracy":
            pred_id = torch.argmax(pred_logits)
            pred = labels[pred_id]
            perf = 1.0 if true_label == pred else 0.0
        elif metric == "correct_prob":
            correct_id = labels.index(true_label)
            perf = torch.softmax(pred_logits, dim=-1)[correct_id].item()
        if countergen.config.verbose >= 4:
            print(f"inp={inp} true_label={true_label} pred={pred} perf={perf}")
        return perf

    return run



from typing import Dict, Iterable, List, Mapping

import torch
from attrs import define
from torch import nn
from countergentorch.editing.activation_utils import get_corresponding_activations

from countergen.types import AugmentedSample, Category

from countergentorch.tools.math_utils import project


@define
class ActivationsDataset(torch.utils.data.Dataset):
    """Dataset of activations with utilities to compute activations and project them."""

    x_data: torch.Tensor  #: 2D float32 tensor of shape (samples, hidden_dimension)
    y_data: torch.Tensor  #: 1D long tensor of shape (samples,) where one number is one category

    @classmethod
    def from_augmented_samples(cls, samples: Iterable[AugmentedSample], model: nn.Module, modules: Iterable[nn.Module]):
        """Compute the activations of the model on the variations of the samples at the output of the given modules.

        The modules are assumed to have outputs of the same shape"""
        activations_dict = get_corresponding_activations(samples, model, modules)
        return ActivationsDataset.from_activations_dict(activations_dict)

    @classmethod
    def from_activations_dict(
        cls, activations: Mapping[Category, List[Dict[nn.Module, torch.Tensor]]], device: str = "cpu"
    ):
        """Group all activations regardless of the module and the sequence position.

        All activations must be of the same shape, (seq_len, hid_size)."""

        x_data_l = []
        y_data_l = []
        for i, li in enumerate(activations.values()):
            # Flatten the activations
            all_activations = []
            for d in li:
                for activation in d.values():
                    all_activations.append(activation)
            x = torch.cat(all_activations)
            x_data_l.append(x)
            y = torch.zeros(x.shape[0], dtype=torch.long)
            y[:] = i
            y_data_l.append(y)
        return ActivationsDataset(torch.cat(x_data_l).to(device), torch.cat(y_data_l).to(device))

    def project(self, dir: torch.Tensor):
        """Return a new dataset where activations have been projected along the dir vector."""
        dir_norm = (dir / torch.linalg.norm(dir)).to(self.x_data.device)
        new_x_data = project(self.x_data, dir_norm[None, :])
        return ActivationsDataset(new_x_data, self.y_data)

    def project_(self, dir: torch.Tensor):
        """Modify activations by projecteding them along the dir vector."""
        dir_norm = (dir / torch.linalg.norm(dir)).to(self.x_data.device)
        self.x_data = project(self.x_data, dir_norm[None, :])

    def __len__(self):
        return len(self.x_data)

    def __getitem__(self, idx):
        x = self.x_data[idx, :]
        y = self.y_data[idx]
        sample = (x, y)
        return sample
